/**************************************************************************
 * Aurora Aries 16000P Serial Multiport Driver (8000P to include)
 *
 * Copyright (c) 1999 Florian Lohoff (flo@rfc822.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **************************************************************************/

/*
 * $VER: 0.72 (990604) Florian Lohoff (flo@rfc822.org)
 *
 * History:
 * 0.72	990604	- Found a buffer wrap bug - Now we can up/download
 * 		  on a serial null-modem and minicom although
 * 		  i still see 3 times the interrupts on receive
 * 		  than on send - it seems i have to implement
 * 		  receive irq threshold and no-new-data-timeout
 * 0.71	990604	- Need ioctl TCSBRK for speeds >=38400 it seems
 * 		  at least for minicom and zmodem up/download
 * 0.7	990604	- Discovered that the Minicom close issue is ok
 *		  though it might be the first working version
 *
 */

/*
 * TODO: (In order of Priority)
 * - Clear SMP/Locking issues (I am not aware of the state)
 * - Software Flow Control
 * - Callout Devices (Are obsolete but someone might need them)
 * - Backport to 2.0.x
 *
 */

#include <linux/module.h>
#include <linux/pci.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/serial.h>
#include <linux/tqueue.h>

#include <asm/param.h>
#include <asm/delay.h>
#include <asm/io.h>
#include <asm/spinlock.h>
#include <asm/uaccess.h>

#include "aurora.h"

#define AA_DEBUG

/*
 *
 * Static structures ... hmmm
 *
 */

static struct aa_board_struct	aa_board[AA_MAX_BOARD];

static unsigned long baud_table[] =  {
	0, 50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800,
	9600, 19200, 38400, 57600, 115200, 0,
};

static struct tty_driver	aa_driver;
static int			aa_refcount;
static struct tty_struct	*aa_table[AA_MAX_PORT];
static struct termios		*aa_termios[AA_MAX_PORT];
static struct termios		*aa_termios_locked[AA_MAX_PORT];


/*
 *
 * Spaghetticode starts here ....
 *
 */
void aa_softint(void *private_) {
	struct port_struct	*port=(struct port_struct *) private_;
	struct tty_struct	*tty;
#ifdef AA_DEBUG
	printk(KERN_INFO "Softint executed\n");
#endif
	if (!(tty=port->tty))
		return;

	if (test_and_clear_bit(RS_EVENT_WRITE_WAKEUP, &port->event)) {
		if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
				tty->ldisc.write_wakeup)
				(tty->ldisc.write_wakeup)(tty);
		wake_up_interruptible(&tty->write_wait);
	}
}

inline void aa_transmit_irq(struct cd1865_struct *cd1865) {
	struct port_struct	*port;
	struct tty_struct	*tty;
	long			ba;
	unsigned char		ack;
	int			i,
				pnum;
#ifdef AA_DEBUG
	printk(KERN_INFO "txirq\n");
#endif

	/* Read Transmit Ack register */
	ack=readb(cd1865->cd1865_base + CD1865_TRAR);

	pnum=readb(cd1865->cd1865_base + CD1865_GSCR1) >> 2;

	/* Get port struct base adr requesting TXRDY request */
	port=&cd1865->port[pnum & 0x7];

	tty=port->tty;
	ba=port->cd1865->cd1865_base;

	writeb(port->port, ba + CD1865_CAR); 

	if (port->srer & CD1865_SRER_TXMPTY) {
		port->srer&=~CD1865_SRER_TXMPTY;
		writeb(port->srer, ba + CD1865_SRER);
		return;
	}

	/* Do we have anything to tansmit or have we been stopped ? */
	if ((port->xmit_cnt <= 0 && !port->break_length) ||
		tty->stopped ||
		tty->hw_stopped) {
#ifdef AA_DEBUG
		printk(KERN_INFO "clr TXRDY #1\n");
#endif
		port->srer &= ~CD1865_SRER_TXRDY;
		writeb(port->srer, ba + CD1865_SRER);
		return;
	}

	/* Do we have to transmit a break ? */
	if (port->break_length) {
		if (port->break_length > 0) {
			/* First time we enter here ? */
			if (port->cor2 & CD1865_COR2_ETC) {
				/* Send Esc Character */
				writeb(CD1865_C_ESC, ba + CD1865_TDR);
				/* Start Break */
				writeb(CD1865_C_SBRK, ba + CD1865_TDR);
				port->cor2&=~CD1865_COR2_ETC;
			}
			/* Send break delay - calculate the maxdelay for 8 bit */
			i=MIN(port->break_length, 0xff);
			/* Send ESC character */
			writeb(CD1865_C_ESC, ba + CD1865_TDR);
			/* Send Delay command */
			writeb(CD1865_C_DELAY, ba + CD1865_TDR);
			/* Send Delay value */
			writeb(i, ba + CD1865_TDR);
			/* Calc new break_length */
			port->break_length-=i;
		}
		/* Until now we have a max of 5 chars in FIFO 
		 * do we want to send an end break now ? */
		if (port->break_length <= 0) {
			/* Ok - we escape the End Break now */
			writeb(CD1865_C_ESC, ba + CD1865_TDR);
			writeb(CD1865_C_EBRK, ba + CD1865_TDR);
			/* End the Embedded Transmit Commands (Command Escaping) */
			writeb(port->cor2, ba + CD1865_COR2);
			/* Send COR2CHNG Channel Command */
			writeb(CD1865_CCR_COR2CHNG, ba + CD1865_CCR);
		}
		/* FIXME: Documentation says we only need to send an
		 * ENDBreak if there are no characters - We could just disable
		 * COR2_ETC and fill the fifo with ordinary data
		 */
		return;
	}

	/* Fill fifo */
	for(i=0;i<CD1865_FIFO_DEPTH && port->xmit_cnt > 0;i++) {
		writeb(port->buffer[port->xmit_tail++], ba + CD1865_TDR);
		if (port->xmit_tail >= AA_PORT_BUFFER_SIZE) 
			port->xmit_tail=0;
		port->xmit_cnt--;
	}

	/* Did we run out of chars ? Clear TXRDY service request */
	if (port->xmit_cnt <= 0) {
#ifdef AA_DEBUG
		printk(KERN_INFO "clr TXRDY #2\n");
#endif
		port->srer &= ~CD1865_SRER_TXRDY;
		writeb(port->srer, ba + CD1865_SRER);
	}

	if (port->xmit_cnt <= port->wakeup_chars) {
		set_bit(RS_EVENT_WRITE_WAKEUP, &port->event);
		queue_task(&port->task, &tq_immediate);
		mark_bh(IMMEDIATE_BH); 
	}
}

inline void aa_receive_exception(struct cd1865_struct *cd1865) {
	struct port_struct	*port;
	struct tty_struct	*tty;
	unsigned long		ba=cd1865->cd1865_base;
	unsigned char		pnum,
				ch,
				status;
#ifdef AA_DEBUG
	printk(KERN_INFO "rxirq (exception)\n");
#endif

	pnum=readb(ba + CD1865_GSCR1) >> 2;

	/* Get port struct base adr requesting TXRDY request */
	port=&cd1865->port[pnum & 0x7];
	tty=port->tty;

	printk(KERN_INFO "port %d\n",port->port);

	/* Get char */
	ch=readb(ba + CD1865_RDR);

	/* Get Receiver Status */
	status = readb(ba + CD1865_RCSR);

	if (!status) 
		return;

	if (status & CD1865_RCSR_TIMEOUT) {
		printk(KERN_INFO "Receiver Timeout\n");
		return;
	} else if (status & CD1865_RCSR_BREAK) {
		*tty->flip.flag_buf_ptr++ = TTY_BREAK;
	} else if (status & CD1865_RCSR_PARITY) {
		*tty->flip.flag_buf_ptr++ = TTY_PARITY;
	} else if (status & CD1865_RCSR_FRAME) {
		*tty->flip.flag_buf_ptr++ = TTY_FRAME;
	} else if (status & CD1865_RCSR_OVERRUN) {
		*tty->flip.flag_buf_ptr++ = TTY_OVERRUN;
	} else 
		*tty->flip.flag_buf_ptr++ = 0;
	
	*tty->flip.char_buf_ptr++ = ch;
	tty->flip.count++;

	queue_task(&tty->flip.tqueue, &tq_timer);
}

inline void aa_receive_data(struct cd1865_struct *cd1865) {
	struct port_struct	*port;
	struct tty_struct	*tty;
	unsigned long		ba=cd1865->cd1865_base;
	unsigned char		count,
				pnum;
	
#ifdef AA_DEBUG
	printk(KERN_INFO "rxirq (data)\n");
#endif

	pnum=readb(ba + CD1865_GSCR1) >> 2;

	/* Get port struct base adr requesting RC request */
	port=&cd1865->port[pnum & 0x7];

	tty=port->tty;

	count=readb(ba + CD1865_RDCR);

	while(count-- && tty->flip.count < TTY_FLIPBUF_SIZE) {
		*tty->flip.char_buf_ptr++ = readb(ba + CD1865_RDR);
		*tty->flip.flag_buf_ptr++ = 0;
		tty->flip.count++;
	}

	queue_task(&tty->flip.tqueue, &tq_timer);
}

static void aa_interrupt(int irq, void *dev_id, struct pt_regs *regs) {
	struct aa_board_struct *board = (struct aa_board_struct *) dev_id;
	int		i;
	unsigned char	status,
			ack;

	for(i=0;i<AA_MAX_CD1865;i++) {
		if (board->cd1865[i].cd1865_base) {
			if ((status = readb(board->cd1865[i].cd1865_base + CD1865_SRSR) & 	
					( CD1865_SRSR_RREQ | CD1865_SRSR_TREQ | CD1865_SRSR_MREQ ))) {
				if (status & CD1865_SRSR_TREQ) {
					aa_transmit_irq(&board->cd1865[i]);
				} else if (status & CD1865_SRSR_RREQ) {
					/* Read Transmit Ack register */
					ack=readb(board->cd1865[i].cd1865_base + CD1865_RRAR);

					if ((ack & 7) == 7) 
						aa_receive_exception(&board->cd1865[i]);
					else if ((ack & 7) == 3) 
						aa_receive_data(&board->cd1865[i]);
				} else if (status & CD1865_SRSR_MREQ) {
					printk(KERN_INFO "  Unimplemented MREQ\n");
				}	

				/* Write to End Of Service Request Register */
				writeb(0, board->cd1865[i].cd1865_base + CD1865_EOSRR);
			}
		}
	}
}


static int aa_write(struct tty_struct *tty, int from_user, const unsigned char *buf, int count) {
	struct port_struct	*port = tty->driver_data;
	unsigned char		*mbuf = 0;
	unsigned long		ba;
	int			i,
				j,
				copied=-EFAULT;

#ifdef AA_DEBUG
	printk(KERN_INFO "aa_write called\n");
#endif
	/* Do we have an tty and a buffer ? */
	if (!tty) 
		return(0);

	/* If we copy from userspace wait until page is free */
	if (from_user)
		down(&port->cd1865->page_sem);
	
	/* Find the min of chars to write an remaining buffer */
	i = MIN(count, AA_PORT_BUFFER_SIZE - port->xmit_cnt - 1);
	
	/* Do we have space ? */
	if (i) {
		/* Should we copy from user ? */
		if (from_user) {
			/* Was the copy successful ? */
			copy_from_user((unsigned char *) port->cd1865->page, buf, i);
			mbuf = (unsigned char *) port->cd1865->page;
		} else
			mbuf=(unsigned char *) buf;
		/* Have we been successful in copy from user or
		 * dont we have to ?
		 */
		if (mbuf) {
			/* Calc size for first copy */
			j=MIN(i, AA_PORT_BUFFER_SIZE - port->xmit_head);
			/* Copy all or at least until end of buffer */
			memcpy(port->buffer + port->xmit_head, mbuf, j);	

			/* Did the previous memcpy include all bytes ? */
			if (i > j) 
				memcpy(port->buffer, mbuf + j, i - j );

			port->xmit_cnt+=i;
			port->xmit_head+=i;

			/* Wrap xmit_head and dont rely on AA_PORT_BUFFER_SIZE = 2^x */
			if (port->xmit_head >= AA_PORT_BUFFER_SIZE) 
				port->xmit_head-=AA_PORT_BUFFER_SIZE;

			copied=i;
		}
	}

	if (from_user)
		up(&port->cd1865->page_sem);
#ifdef AA_DEBUG
	if (port->srer & CD1865_SRER_TXRDY)
		printk(KERN_INFO "aa_write: TXRDY already set\n");
#endif

	if (port->xmit_cnt && !tty->stopped && !tty->hw_stopped &&
		!(port->srer & CD1865_SRER_TXRDY)) {

		ba=port->cd1865->cd1865_base;

		writeb(port->port, ba + CD1865_CAR);

		port->srer |= CD1865_SRER_TXRDY;
		writeb(port->srer, ba + CD1865_SRER);
	}

	return(copied);
}

static void aa_put_char(struct tty_struct * tty, unsigned char ch) {
	struct port_struct	*port = tty->driver_data;
	unsigned long		cpu_flags;
#ifdef AA_DEBUG
	printk(KERN_INFO "put_char called\n");
#endif
	if (!tty)
		return;

	spin_lock_irqsave(&port->buffer_lock, cpu_flags);

	if (port->xmit_cnt >= AA_PORT_BUFFER_SIZE - 1) 
		return;

	port->buffer[port->xmit_head++]=ch;

	if (port->xmit_head >= AA_PORT_BUFFER_SIZE) 
		port->xmit_head=0;

	port->xmit_cnt++;

	spin_unlock_irqrestore(&port->buffer_lock, cpu_flags);
}

static void aa_flush_chars(struct tty_struct * tty) {
	struct port_struct	*port = tty->driver_data;
	long			ba = port->cd1865->cd1865_base;
	unsigned long		cpu_flags;

	/* No bytes, no buffer, or stopped or already enabled */
	if (port->xmit_cnt <= 0 || tty->stopped ||
		tty->hw_stopped || (port->srer & CD1865_SRER_TXRDY)) 
		return;

	/* Lock cd1865 */
	spin_lock_irqsave(&port->cd1865->access_lock, cpu_flags);

	/* Enable Transmit Ready Service Request */
	port->srer|=CD1865_SRER_TXRDY;

	/* Select port */
	writeb(port->port, ba + CD1865_CAR);
	/* Write Service Request Enable Register */
	writeb(port->srer, ba + CD1865_SRER);

	/* Unlock cd1865 */
	spin_unlock_irqrestore(&port->cd1865->access_lock, cpu_flags);
}

static int aa_write_room(struct tty_struct *tty) {
	struct port_struct	*port=tty->driver_data;
	return MAX(0, AA_PORT_BUFFER_SIZE - port->xmit_cnt - 1);
}

static int aa_chars_in_buffer(struct tty_struct *tty) {
	struct port_struct	*port=tty->driver_data;
	return(port->xmit_cnt);
}

static void aa_flush_buffer(struct tty_struct *tty) {
	struct port_struct	*port=tty->driver_data;
	unsigned long		cpu_flags;
#ifdef AA_DEBUG	
	printk(KERN_INFO "aa_flush_buffer called\n");
#endif
	spin_lock_irqsave(&port->cd1865->access_lock, cpu_flags);
	port->xmit_cnt=port->xmit_head=port->xmit_tail=0;
	spin_unlock_irqrestore(&port->cd1865->access_lock, cpu_flags);

	wake_up_interruptible(&tty->write_wait);

	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
			tty->ldisc.write_wakeup)
		(tty->ldisc.write_wakeup)(tty);
}

static int aa_get_modem_info(struct port_struct *port, unsigned int *value) {
	unsigned long	cpu_flags;
	unsigned long	ba = port->cd1865->cd1865_base;
	unsigned char	status;
	unsigned int	result;

	spin_lock_irqsave(&port->cd1865->access_lock, cpu_flags);

	writeb(port->port, ba + CD1865_CAR);
	status=readb(ba + CD1865_MSVR);	

	spin_unlock_irqrestore(&port->cd1865->access_lock, cpu_flags);

	if (C_CRTSCTS(port->tty)) 
		result = ((status & CD1865_MSVR_DTR) ? TIOCM_RTS : 0) |
			((status & CD1865_MSVR_CD)  ? TIOCM_CAR : 0) |
			((status & CD1865_MSVR_CTS) ? TIOCM_CTS : 0);
	else
		result = ((status & CD1865_MSVR_DTR) ? TIOCM_DTR : 0) |
			((status & CD1865_MSVR_CD)  ? TIOCM_CAR : 0) |
			((status & CD1865_MSVR_CTS) ? TIOCM_CTS : 0);
#ifdef AA_DEBUG
	if (status & CD1865_MSVR_CD) {
		printk(KERN_INFO "Carrier detect on ?\n");
	}
#endif
	put_user(result,(unsigned long *) value);
	return(0);
}

static int aa_set_modem_info(struct port_struct *port, unsigned int cmd, unsigned int *value) {
	int		error;
	unsigned long	cpu_flags,
			arg,
			ba = port->cd1865->cd1865_base;

	error=verify_area(VERIFY_READ, value, sizeof(int));
	if (error)
		return error;

	get_user(arg, (unsigned long *) value);

	switch (cmd) {
	case TIOCMBIS: 
		if (C_CRTSCTS(port->tty)) {
			if (arg & TIOCM_RTS)
				port->msvr |= CD1865_MSVR_DTR; 
		} else {
			if (arg & TIOCM_DTR)
				port->msvr |= CD1865_MSVR_DTR; 
		}	     
		break;
	case TIOCMBIC:
		if (C_CRTSCTS(port->tty)) {
			if (arg & TIOCM_RTS)
				port->msvr &= ~CD1865_MSVR_DTR;
		} else {
			if (arg & TIOCM_DTR)
				port->msvr &= ~CD1865_MSVR_DTR;
		}
		break;
	case TIOCMSET:
		if (C_CRTSCTS(port->tty)) {
	  		port->msvr = (arg & TIOCM_RTS) ? 
			                         (port->msvr |  CD1865_MSVR_DTR) : 
			                         (port->msvr & ~CD1865_MSVR_DTR);
		} else {
			port->msvr = (arg & TIOCM_DTR) ?
			                         (port->msvr |  CD1865_MSVR_DTR):
			                         (port->msvr & ~CD1865_MSVR_DTR);
		}
		break;
	default:
		return -EINVAL;
	}

	spin_lock_irqsave(&port->cd1865->access_lock, cpu_flags);

	writeb(port->port, ba + CD1865_CAR);
	writeb(port->msvr, ba + CD1865_MSVR);

	spin_unlock_irqrestore(&port->cd1865->access_lock, cpu_flags);

	return 0;
}

inline void aa_send_break(struct port_struct *port, unsigned short length) {
	unsigned long	cpu_flags,
			ba=port->cd1865->cd1865_base;
	int		i;

	spin_lock_irqsave(&port->cd1865->access_lock,cpu_flags);

	/* Set break length for txirq */
	port->break_length=AA_TPS / HZ * length;

	/* Enable Embedded Transmitter Command (Escaped Special Chars) */
	port->cor2|=CD1865_COR2_ETC;
	port->srer|=CD1865_SRER_TXRDY;

	writeb(port->port, ba + CD1865_CAR);
	writeb(port->cor2, ba + CD1865_COR2);
	writeb(port->srer, ba + CD1865_SRER);

	writeb(CD1865_CCR_COR2CHNG, ba + CD1865_CCR);

	for(i=0;i<AA_CCR_TIMEOUT && readb(ba + CD1865_CCR) != 0x0;i++);

	/* Didnt CCR zero ? */
	if (i == AA_CCR_TIMEOUT)
		printk(KERN_ERR "Ops - CCR didnt went to ready state\n");

	spin_unlock_irqrestore(&port->cd1865->access_lock,cpu_flags);
}

static int aa_ioctl(struct tty_struct *tty, struct file *fp, unsigned int cmd, unsigned long arg) {
	struct port_struct	*port = tty->driver_data; 
	int			error;

	switch(cmd) {
		case TCSBRK:
			error = tty_check_change(tty);
			if (error)
				return error;
			tty_wait_until_sent(tty, 0);
			if (!arg)
				aa_send_break(port, HZ/4);      /* 1/4 second */
			return(0);
		case TCSBRKP:
			printk(KERN_ERR "aa_ioctl: unimplemented TCSBRKP\n");
			return(0);
		case TIOCGSOFTCAR:
			printk(KERN_ERR "aa_ioctl: unimplemented TIOCGSOFTCAR\n");
			return(0);
		case TIOCSSOFTCAR:
			printk(KERN_ERR "aa_ioctl: unimplemented TIOCSSOFTCAR\n");
			return(0);
		case TIOCMGET:
			error=verify_area(VERIFY_WRITE, (void *) arg, sizeof(unsigned int ));
			if (error)
				return(error);
			return aa_get_modem_info(port, (unsigned int *) arg);
		case TIOCMBIS:
		case TIOCMBIC:
		case TIOCMSET:
			return aa_set_modem_info(port, cmd, (unsigned int *) arg);
		case TIOCGSERIAL:
			printk(KERN_ERR "aa_ioctl: unimplemented TIOCGSERIAL\n");
			return(0);
		case TIOCSSERIAL:
			printk(KERN_ERR "aa_ioctl: unimplemented TIOCSSERIAL\n");
			return(0);
		default:
			return -ENOIOCTLCMD;
	}

	return(0);
}

static void aa_throttle(struct tty_struct *tty) {
	printk(KERN_ERR "unimplemented aa_throttle called\n");
}

static void aa_unthrottle(struct tty_struct * tty) {
	printk(KERN_ERR "unimplemented aa_unthrottle called\n");
}

static void aa_stop(struct tty_struct * tty) {
	printk(KERN_ERR "unimplemented aa_stop called\n");
}

static void aa_start(struct tty_struct * tty) {
	struct port_struct	*port=tty->driver_data;
	unsigned long		ba=port->cd1865->cd1865_base,
				cpu_flags;

	/* Lock cd1865 */
	spin_lock_irqsave(&port->cd1865->access_lock, cpu_flags);

	/* Enable Transmit Ready Service Request */
	port->srer|=CD1865_SRER_TXRDY;

	/* Select port */
	writeb(port->port, ba + CD1865_CAR);
	/* Write Service Request Enable Register */
	writeb(port->srer, ba + CD1865_SRER);

	/* Unlock cd1865 */
	spin_unlock_irqrestore(&port->cd1865->access_lock, cpu_flags);
}

static void aa_hangup(struct tty_struct * tty) {
	printk(KERN_ERR "unimplemented aa_hangup called\n");
}


void aa_release_driver(void ) {
	tty_unregister_driver(&aa_driver);
#ifdef AA_DEBUG
	printk(KERN_INFO "unregistered aurora driver\n");
#endif
}

void aa_shut_port(struct port_struct *port) {
	long		ba=port->cd1865->cd1865_base;
	unsigned long	cpu_flags;

	spin_lock_irqsave(&port->cd1865->access_lock, cpu_flags);

	/* Select correct port */
	writeb(port->port,CD1865_CAR);

	/* Clear Service Request */
	port->srer=0;
	
	/* Write to chip */
	writeb(port->srer, ba + CD1865_SRER);

	/* Clear DTR Bit */
	port->msvr&=~CD1865_MSVR_DTR;

	/* Write Modem Lines (Drop DTR) */
	writeb(port->msvr, ba + CD1865_MSVR);

	spin_unlock_irqrestore(&port->cd1865->access_lock, cpu_flags);
}

static void aa_close(struct tty_struct *tty, struct file *fp) {
	struct port_struct *port=tty->driver_data;

	/* Port NULL ? */
	if (!port) {
		printk(KERN_ERR "Oops - port in tty struct is NULL\n");
		return;
	}

	tty->closing=1;

	/* Wait until sent - Will be woken up by BH handler */
	if (port->closing_wait)
		tty_wait_until_sent(tty, port->closing_wait);

	/* Reset port - tell to shut up */
	aa_shut_port(port);

	/* Wake up the possibly opening process */
	wake_up_interruptible(&port->close_wait);

	port->count--;

	MOD_DEC_USE_COUNT;
}

static void aa_set_speed(struct port_struct *port) {
	int	baud,i;
	long	ba=port->cd1865->cd1865_base;
	unsigned long	cpu_flags;

	/* Get the right speed */
	baud=C_BAUD(port->tty);

#if 0	/* Dont know about this - comment it out */
	if (baud & CBAUDEX) {
		baud &= ~CBAUDEX;
		if (baud < 1 || baud > 2)
			port->tty->termios->c_cflag &= ~CBAUDEX;
		else
			baud += 15;
	}
#endif 

	/* OERGS we have flags to select speed above 38400 ... */
	if (baud == 15) {
		if ((port->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI)
			baud ++;
		if ((port->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI)
			baud += 2;
	}

	baud=baud_table[baud];

	/* If Baud = 0 drop DTR */
	if (!baud) {
		if (!C_CRTSCTS(port->tty)) {
			port->msvr &= ~CD1865_MSVR_DTR;

			spin_lock_irqsave(&port->cd1865->access_lock, cpu_flags);

			writeb(port->port, ba + CD1865_CAR);
			writeb(port->msvr, ba + CD1865_MSVR);

			spin_unlock_irqrestore(&port->cd1865->access_lock, cpu_flags);
		}
		return;
	} else if (!C_CRTSCTS(port->tty)) 
		port->msvr |= CD1865_MSVR_DTR;

	/* calculate baud rate */
	port->bpr=(( AA_OSCFREQ + baud/2 ) / baud + AA_TPC/2 ) / AA_TPC;

	/* Calculate wakeup_chars (baud/hz*2) between 0 and AA_PORT_BUFFER_SIZE  */(
	port->wakeup_chars = MIN(MAX(0,((baud + HZ/2) / HZ) * 2 - CD1865_FIFO_DEPTH), AA_PORT_BUFFER_SIZE - 1));

	/* Clear Channel Option Register 1 (No Bits, Stop Bits, Parity) */
	port->cor1=0;

	/* Set No Bits */
	switch (C_CSIZE(port->tty)) {
		case CS5:
			port->cor1|=CD1865_COR1_5BITS;
			break;
		case CS6:
			port->cor1|=CD1865_COR1_6BITS;
			break;
		case CS7:
			port->cor1|=CD1865_COR1_7BITS;
			break;
		case CS8:
			port->cor1|=CD1865_COR1_8BITS;
	}

	/* We dont do 1 1/2 or 2 1/2 so just accept 2 */
	if (C_CSTOPB(port->tty))
		port->cor1|=CD1865_COR1_STOP2;

	/* Lets first ignore parity */
	port->cor1 |= CD1865_COR1_PARIGNORE;

	/* Do we want to check parity ? */
	if (C_PARENB(port->tty)) {
		port->cor1 |= CD1865_COR1_PARNORM;
		/* odd parity ? */
		if (C_PARODD(port->tty))
			port->cor1 |= CD1865_COR1_PARODD;
		/* Should we check parity ? -> reset IGNORE*/
		if (I_INPCK(port->tty))
			port->cor1 &= ~CD1865_COR1_PARIGNORE;
	}

	/* Hardware flow control ? */
	if (C_CRTSCTS(port->tty)) 
		port->cor2|=CD1865_COR2_CTSAE;

	/* FIXME Software Flow Control ? */

	if (!C_CLOCAL(port->tty)) {
		/* Service Request Enable CD */
		port->srer|=CD1865_SRER_CD;
		/* Report h/l transitions */
		port->mcor1|=CD1865_MCOR1_CDZD;	
		/* Report l/h transitions */
		port->mcor2|=CD1865_MCOR2_CDOD;
	}

	if (C_CREAD(port->tty))  {
#ifdef AA_DEBUG
		printk(KERN_INFO "Enabling receive irq\n");
#endif
		port->srer|=CD1865_SRER_RDSR;
	}

	/*
	 * Lock it - we will be accesing the Serial port
	 * channel specific registers
	 *
	 */
	spin_lock_irqsave(&port->cd1865->access_lock,cpu_flags);

	/* Select port on chip */
	writeb(port->port, ba + CD1865_CAR);

	/* Set Tx baud rate */
	writeb(port->bpr >> 8, ba + CD1865_TBPRH);
	writeb(port->bpr & 0xff, ba + CD1865_TBPRL);

	/* Set Rx baud rate */
	writeb(port->bpr >> 8  , ba + CD1865_RBPRH);
	writeb(port->bpr & 0xff, ba + CD1865_RBPRL);

	/* Write Channel Option Register 1 */
	writeb(port->cor1, ba + CD1865_COR1);

	/* Write Channel Option Register 2 */
	writeb(port->cor2, ba + CD1865_COR2);

	/* Write Channel Option Register 3 */
	writeb(port->cor3, ba + CD1865_COR3);

	/* Signal cor1/2/3 changes */
	writeb(CD1865_CCR_CORCHNGC | CD1865_CCR_COR1CHNG | CD1865_CCR_COR2CHNG | CD1865_CCR_COR3CHNG,
			ba + CD1865_CCR);

	for(i=0;i<AA_CCR_TIMEOUT && readb(ba + CD1865_CCR) != 0x0;i++);

	/* Didnt CCR zero ? */
	if (i == AA_CCR_TIMEOUT)
		printk(KERN_ERR "Ops - CCR didnt went to ready state\n");

	/* Enable Transmitter & Receiver */
	writeb(CD1865_CCR_CCC | CD1865_CCR_TXENABLE | CD1865_CCR_RXENABLE, ba + CD1865_CCR);

	/* Enable Service Requests */
	writeb(port->srer, ba + CD1865_SRER);

	/* Set DTR active */
	writeb(port->msvr, ba + CD1865_MSVR);

	/* Ok - unlock */
	spin_unlock_irqrestore(&port->cd1865->access_lock, cpu_flags);

	return;
}

static void aa_set_termios(struct tty_struct *tty, struct termios *old_termios) {
	struct port_struct	*port=tty->driver_data;

	if (tty->termios->c_cflag == old_termios->c_cflag &&
		tty->termios->c_iflag == old_termios->c_iflag)
		return;

	aa_set_speed(port);

	if ((old_termios->c_cflag & CRTSCTS) &&
		!(tty->termios->c_cflag & CRTSCTS)) {
		tty->hw_stopped=0;
		aa_start(tty);
	}
}

static int aa_open(struct tty_struct *tty, struct file *fp) {
	struct port_struct	*port = 0;
	int			i, j,
				portcnt = 0,
				portno = MINOR(tty->device) - tty->driver.minor_start;

	/* Board loop */
	for(i=0;i<AA_MAX_BOARD && !(port);i++) {
		/* loop count of cd1865 and cd1865 found */
		for(j=0;j<AA_MAX_CD1865 &&
			!(port) &&
			aa_board[i].cd1865[j].cd1865_base;j++,portcnt+=AA_CD1865_PORT) {

			/* Is it this CD1865 ? */
			if ((portcnt+AA_CD1865_PORT) > portno) {

				/* Did this board succeed reset ? */
				if (aa_board[i].cd1865[j].state & AA_CD1865_FAILED) 
					return -ENODEV;

				port = &aa_board[i].cd1865[j].port[portno % AA_CD1865_PORT];	
			}
		}
	}
	/* Did we find a port ? */
	if (!(port)) {
		printk(KERN_ERR "No such device\n");
		return -ENODEV;
	}
#ifdef AA_DEBUG
	printk(KERN_INFO "Opened board %d cd1865 %d port %d\n",
			i,j,port->port);
#endif
	/* Wait for closing of port */
        if (tty_hung_up_p(fp) || port->flags & ASYNC_CLOSING) 
		interruptible_sleep_on(&port->close_wait);

	/* Do we have a buffer ? */
	if (!port->buffer) 
		if (!(port->buffer=(unsigned char *) get_free_page(GFP_KERNEL))) {
			printk(KERN_ERR "ops - no memory for aa_open\n");
			return -ENOMEM;
		}

	/* Store the pointers over cross*/
	tty->driver_data=port;
	port->tty=tty;

	/* FIXME bail out if closing */
	/* FIXME block_til_ready only for interaction with callout devices ? */
	/* FIXME set termios */

	aa_set_speed(port);

	port->session = current->session;
	port->pgrp = current->pgrp;

	port->count++;

	MOD_INC_USE_COUNT;

	return(0);
}


int aa_setup_driver(void ) {
	int error;

	/*
	 *
	 * Initializing driver
	 *
	 */

	memset(&aa_driver, 0, sizeof(aa_driver));	

	aa_driver.magic			=	TTY_DRIVER_MAGIC;
	aa_driver.name			=	"ttyA";
	aa_driver.major			=	AA_MAJOR;
	aa_driver.minor_start		=	AA_MINOR;
	aa_driver.num			=	AA_MAX_PORT;
	aa_driver.type			=	TTY_DRIVER_TYPE_SERIAL;
	aa_driver.subtype		=	1;
	aa_driver.init_termios		=	tty_std_termios;
	aa_driver.init_termios.c_cflag	=	B9600 | CS8 | CREAD | HUPCL | CLOCAL;
	aa_driver.flags			= 	TTY_DRIVER_REAL_RAW;
	aa_driver.refcount		=	&aa_refcount;
	aa_driver.table			=	aa_table;
	aa_driver.termios		=	aa_termios;
	aa_driver.termios_locked	=	aa_termios_locked;
	
	aa_driver.open			=	aa_open;
	aa_driver.close			= 	aa_close;
	aa_driver.write			=	aa_write;
	aa_driver.put_char		=	aa_put_char;
	aa_driver.flush_chars		=	aa_flush_chars;
	aa_driver.write_room		=	aa_write_room;
	aa_driver.chars_in_buffer	=	aa_chars_in_buffer;
	aa_driver.flush_buffer		=	aa_flush_buffer;
	aa_driver.ioctl			=	aa_ioctl;
	aa_driver.throttle		=	aa_throttle;
	aa_driver.unthrottle		=	aa_unthrottle;
	aa_driver.set_termios		=	aa_set_termios;
	aa_driver.stop			=	aa_stop;
	aa_driver.start			=	aa_start;
	aa_driver.hangup		=	aa_hangup;

	if ((error=tty_register_driver(&aa_driver))) {
		printk(KERN_ERR "aa: Could not register Aries Driver error = %d\n", error);
		return(1);
	}
#ifdef AA_DEBUG
	printk(KERN_INFO "registered aurora driver\n");
#endif
	return(0);
}

int aa_reset_cd1865(struct cd1865_struct *cd1865) {
	long		ba = cd1865->cd1865_base;
	unsigned long	cpu_flags;
	int		i;

	/* Do we need to lock ? */
	spin_lock_irqsave(&cd1865->access_lock, cpu_flags);

	/* Set GSVR to 0x0 */
	writeb(0x00, ba + CD1865_GSVR);
	for(i=0;i<AA_CCR_TIMEOUT && readb(ba + CD1865_CCR) != 0x0;i++);

	/* Didnt CCR zero ? */
	if (i == AA_CCR_TIMEOUT) 
		return(1);
		
	/* Reset */
	writeb(CD1865_CCR_HWRESET, ba + CD1865_CCR);
	for(i=0;i<AA_HWRESET_TIMEOUT && readb(ba + CD1865_GSVR) != 0xff;i++);

	/* Did the chip get ready after reset ? */
	if (i == AA_HWRESET_TIMEOUT) 
		return(1);

	/*
	 *
	 * Set values for register access based Service Request ack
	 *
	 */

	writeb(CD1865_MRAR, ba + CD1865_MSMR);		/* Modem Service */
	writeb(CD1865_TRAR, ba + CD1865_TSMR);		/* Transmit Service */
	writeb(CD1865_RRAR, ba + CD1865_RSMR);		/* Receive Service */

	/* Set RegAckEn */
	writeb(readb(ba + CD1865_SRCR) | CD1865_SRCR_REGACKEN, ba + CD1865_SRCR);

	/* Set prescaler for clocking */
	writeb(AA_PRESCALER >> 8, ba + CD1865_PPRH);
	writeb(AA_PRESCALER & 0xff, ba + CD1865_PPRL);

	spin_unlock_irqrestore(&cd1865->access_lock, cpu_flags);
		
	return(0);
}

int aa_find_cards(void ) {
	unsigned char		pci_bus,
				pci_device_fn;
	struct pci_dev		*pdev;
	struct port_struct	*port;
	static int		pci_index;
	int			cb=0,
				cd1865_no,
				irq,
				i, j,
				ports_found=0;
	long			port_base,
				cd1865_base,
				plx_base,
				ba;

	memset(&aa_board,0,sizeof(aa_board));

	/* Do we have an PCI Bios *Oergs* */
	if (! pcibios_present()) 
		return -ENODEV;

	/* Try to setup the char device driver */
	if (aa_setup_driver()) 
		return -EIO;

	/*
	 *
	 * Try finding Aries 16000P board
	 *
	 */
	for (pci_index=0;pci_index < AA_MAX_BOARD; pci_index++ ) {
		if (!(pcibios_find_device(PCI_VENDOR_ID_AURORA,
				PCI_DEVICE_ID_AURORA_16000P,
				pci_index, &pci_bus,
				&pci_device_fn))) {

			/* Got one ... get *pci device struct */
			pdev = pci_find_slot(pci_bus, pci_device_fn);

			/*
			 *
			 * Get all infos from the pci dev struct
			 *
			 */
			plx_base=(ulong) ioremap(pdev->base_address[0] &
							PCI_BASE_ADDRESS_MEM_MASK,
						AA_PLX_PCI_SIZE);

			cd1865_base=(ulong) ioremap(pdev->base_address[2] &
							PCI_BASE_ADDRESS_MEM_MASK,
						AA_CHIP_PCI_SIZE);

			port_base=pdev->base_address[1] &
						PCI_BASE_ADDRESS_MEM_MASK;	

			/* We found an 16000P so we have 2xCD1865 */
			cd1865_no=2;

			irq=pdev->irq;

			printk(KERN_INFO "Found Aries 16000P IRQ %d\n", irq);
				
		} else {
			/* FIXME We should now search for 8000P boards
			 * Could please someone fill this gap or send
			 * me an Aries 8000P Board 
			 */
			break;
		} 

		/* Store PCI and port addresses */
		aa_board[cb].plx_base=plx_base;
		aa_board[cb].port_base=port_base;

		/* Get interrupt */
		if (request_irq(irq, aa_interrupt, SA_INTERRUPT, "aurora aries", &aa_board[cb])) {
			printk(KERN_WARNING "  failed to allocate interrupt\n");
			break;
		}

		/* Set interrupt only when it got allocated */
		aa_board[cb].irq=irq;

		printk(KERN_INFO "  Enabling PLX9060 interrupts\n");

		/* Enable PCI Chip interrupts */
		outw(inw(aa_board[cb].port_base + PLX9060_ICSR) | PLX9060_ICSR_PCIENABLE | PLX9060_ICSR_PCILOCAL,
			aa_board[cb].port_base + PLX9060_ICSR);

		/* Loop on found CD1865 and reset them - if succesfull add to cd1865 table */
		for(i=0;i<cd1865_no;i++) {
			/* Base address */
			ba=cd1865_base + i * AA_CD1865_MEM_SIZE;
			/* Store base address */
			aa_board[cb].cd1865[i].cd1865_base=ba;
			/* Store board pointer */
			aa_board[cb].cd1865[i].board=&aa_board[cb];
			/* Initialize spin_lock */
			aa_board[cb].cd1865[i].access_lock = SPIN_LOCK_UNLOCKED;
			/* Initialize bounce buffer semaphore */
			aa_board[cb].cd1865[i].page_sem = MUTEX;
			/* Reset cd1865 */
			if (aa_reset_cd1865(&aa_board[cb].cd1865[i]))  {
				printk(KERN_WARNING "  cd1865 at %08lx reset FAILED !!\n", ba);
				/* Remember the failed reset */
				aa_board[cb].cd1865[i].state|=AA_CD1865_FAILED;
			} else {
				printk(KERN_INFO "  cd1865 at %08lx reset successful\n", ba);
				/* Get Buffer for the Serial Ports (one page per 8 serials) */
				aa_board[cb].cd1865[i].page=get_free_page(GFP_KERNEL);	
				/* Did we get memory ? */
				if (!aa_board[cb].cd1865[i].page) 
					printk(KERN_ERR "  No memory for buffers\n");
				else
					/* Process ports */
					for(j=0;j<AA_CD1865_PORT;j++) {
						port=&aa_board[cb].cd1865[i].port[j];

						/* Set board pointer to all ports */
						port->cd1865=&aa_board[cb].cd1865[i];
						port->port=j;
						port->closing_wait=30*HZ;

						port->task.sync = 0;
						port->task.routine=aa_softint;
						port->task.data=port;

						port->termios = aa_driver.init_termios;

						ports_found++;
					}
			} /* aa_reset_cd1865(ba) */
		} /* CD1865 loop */
	} /* PCI loop */

	if (!(ports_found)) 
		aa_release_driver();

	return(ports_found);
}

void aa_free_resources(void ) {
	int	i,j,k;

	/* Loop on boards */
	for(i=0;i<AA_MAX_BOARD;i++) {

		/* Did we allocate an irq for this board ? */
		if (aa_board[i].irq) {
			/* Release irq for this board */
			free_irq(aa_board[i].irq, &aa_board[i]);
			/* Clear IRQ - be on the safe side */
			aa_board[i].irq=0;
		}

		/* Loop on chips and search for memory allocated */
		for(j=0;j<AA_MAX_CD1865;j++) {
			/* Reset chip (should also stop irqs */
			aa_reset_cd1865(&aa_board[i].cd1865[j]);

			if (aa_board[i].cd1865[j].page) {
				free_page(aa_board[i].cd1865[j].page);
				aa_board[i].cd1865[j].page=0;
			}
			for(k=0;k<AA_CD1865_PORT;k++) {
				if (aa_board[i].cd1865[j].port[k].buffer) {
					free_page((unsigned long) aa_board[i].cd1865[j].port[k].buffer);
					aa_board[i].cd1865[j].port[k].buffer=0;
				}
			}
		}
	}
}

#ifdef MODULE

int init_module(void ) {
	int cards_found;

	cards_found=aa_find_cards();

	if (cards_found) 
		return(0);

	aa_free_resources();
	return -ENODEV;
}

int cleanup_module(void ) {

	aa_release_driver();
	aa_free_resources();

	return(0);
}

#endif /* MODULE */
