
#include <asm/semaphore.h>
#include <asm/spinlock.h>
#include <linux/interrupt.h>

#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif

#define PLX9060_ICSR		0x68		/* Interrupt Control/Status Register */
#define PLX9060_ICSR_PCIENABLE	1<<8
#define	PLX9060_ICSR_PCILOCAL	1<<11

#define PCI_VENDOR_ID_AURORA 0x10b5
#define PCI_DEVICE_ID_AURORA_16000P 0x906d

#define AA_MAJOR 75
#define AA_MINOR 0

#define AA_PLX_PCI_SIZE		256
#define AA_CHIP_PCI_SIZE	1024
#define AA_CD1865_MEM_SIZE	0x100

#define AA_TPS			4000		/* 4 Ticks per 1ms */
#define AA_TPC			16		/* Ticks per character */
#define AA_OSCFREQ		66000000/2	/* 66 Mhz Crystal / 2*/

#define AA_PRESCALER		AA_OSCFREQ / AA_TPS
/*
 *
 * Global Registers
 *
 */
#define CD1865_GFRCR	0x6b	/* Global Firmware Revision Code Rehister */
#define CD1865_SRCR	0x66	/* Service Request Configuration Register */
#define CD1865_PPRH	0x70	/* Prescaler Period Register High */
#define CD1865_PPRL	0x71	/* Prescaler Period Rehister Low */
#define CD1865_MSMR	0x61	/* Modem Service Match Register */
#define	CD1865_TSMR	0x62	/* Transmit Service Match Register */
#define CD1865_RSMR	0x63	/* Receive Service Match Register */
#define CD1865_GSVR	0x40	/* Global Service Vector Register */
#define CD1865_SRSR	0x65	/* Service Request Status Register */
#define CD1865_MRAR	0x75	/* Modem Request Acknowledge Register */
#define CD1865_TRAR	0x76	/* Transmit Request Acknowledge Register */
#define CD1865_RRAR	0x77	/* Receive Request Acknowledge Register */
#define CD1865_GSCR1	0x41	/* Global Service Channel Register 1 */
#define CD1865_GSCR2	0x42	/* Global Service Channel Register 2 */
#define CD1865_GSCR3	0x43	/* Global Service Channel Register 3 */
#define CD1865_CAR	0x64	/* Channel Access Register */
/*
 *
 * Index Indirect Registers
 *
 */
#define CD1865_RDCR	0x07	/* Receive Data Count Register */
#define CD1865_RDR	0x78	/* Receive Data Register */
#define CD1865_RCSR	0x7a	/* Receive Character Status Register */
#define CD1865_TDR	0x7b	/* Transmit Data Register */
#define CD1865_EOSRR	0x7f	/* End Of Service Request Register */
/*
 *
 * Channel Registers
 *
 */
#define CD1865_SRER	0x02	/* Service Request Enable Register */
#define CD1865_CCR	0x01	/* Channel Command Register */
#define CD1865_COR1	0x03	/* Channel Option Register 1 */
#define CD1865_COR2	0x04	/* Channel Option Register 2 */
#define CD1865_COR3	0x05	/* Channel Option Register 3 */
#define CD1865_CCSR	0x06	/* Channel Control Status Register */
#define CD1865_RBR	0x33	/* Receive Bit Register */
#define CD1865_RTPR	0x18	/* Receive Time-Out Period Register */
#define CD1865_RBPRH	0x31	/* Receive Bit Rate Period Register High */
#define CD1865_RBPRL	0x32	/* Receive Bit Rate Period Register Low */
#define CD1865_TBPRH	0x39	/* Transmit Bit Rate Period Register High */
#define CD1865_TBPRL	0x3a	/* Transmit Bit Rate Period Register Low */
#define CD1865_SCHR1	0x09	/* Special Charachter Register 1 */
#define CD1865_SCHR2	0x0a	/* Special Charachter Register 2 */
#define CD1865_SCHR3	0x0b	/* Special Charachter Register 3 */
#define CD1865_SCHR4	0x0c	/* Special Charachter Register 4 */
#define CD1865_MCR	0x12	/* Modem Change Register */
#define CD1865_MCOR1	0x10	/* Modem Change Option Register 1 */
#define CD1865_MCOR2	0x11	/* Modem Change Option Register 2 */
#define	CD1865_MSVR	0x28	/* Modem Signal Value Register */
#define CD1865_MSVRTS	0x29	/* Modem Signal Value Request To Send */
#define CD1865_MSVDTR	0x2a	/* Modem Signal Value Data Terminal Ready */
/*
 *
 * Values
 *
 */
#define CD1865_CCR_HWRESET	0x81

#define CD1865_CCR_CCC		1<<4	/* Channel Control Command */
#define CD1865_CCR_TXENABLE	1<<3
#define CD1865_CCR_TXDISABLE	1<<2
#define CD1865_CCR_RXENABLE	1<<1
#define CD1865_CCR_RXDISABLE	1<<0

#define CD1865_CCR_CORCHNGC	1<<6	/* Chaannel Option Register Change */
#define CD1865_CCR_COR1CHNG	1<<1
#define CD1865_CCR_COR2CHNG	1<<1
#define CD1865_CCR_COR3CHNG	1<<1

#define CD1865_MSVR_RTS		1<<0
#define CD1865_MSVR_DTR		1<<1
#define CD1865_MSVR_CTS		1<<5
#define CD1865_MSVR_CD		1<<6
#define CD1865_MSVR_DSR		1<<7

#define CD1865_COR1_5BITS	0
#define CD1865_COR1_6BITS	1
#define CD1865_COR1_7BITS	2
#define CD1865_COR1_8BITS	3
#define CD1865_COR1_STOP2	1<<3
#define CD1865_COR1_PARIGNORE	1<<4
#define CD1865_COR1_PARFORCE	1<<5
#define CD1865_COR1_PARNORM	1<<6
#define CD1865_COR1_PARODD	1<<7

#define CD1865_COR2_CSRAE	1<<0	/* DSR Automatic Enable */
#define CD1865_COR2_CTSAE	1<<1	/* CTS Automatic Enable */
#define CD1865_COR2_RTSAO	1<<2	/* RTS Automatic Enable */
#define CD1865_COR2_RLM		1<<3	/* Remote Loopback Mode */
#define CD1865_COR2_LLM		1<<4	/* Local Loopback Mode */
#define CD1865_COR2_ETC		1<<5	/* Embedded Transmitter Command Enable */
#define CD1865_COR2_TXLBE	1<<6	/* Transmit in-Band (Xon/Xoff) Flow Control Automatic Enable */
#define CD1865_COR2_IXM		1<<7	/* Implied Xon Mode */

#define CD1865_COR3_SCDE	1<<4	/* Special Character Detection Enable */
#define CD1865_COR3_FCT		1<<5	/* Flow Control Transparency */
#define CD1865_COR3_XOFFCD	1<<6	/* Xoff Character Definition */
#define CD1865_COR3_XONCD	1<<7	/* Xon Character Definition */

#define CD1865_MCOR1_CTSZD	1<<5	/* Detect h/l transition on CTS */
#define CD1865_MCOR1_CDZD	1<<6	/* Detect h/l transition on CD */
#define CD1865_MCOR1_DSRZD	1<<7	/* Detect h/l transition on DSR */

#define CD1865_MCOR2_CTSOD	1<<5	/* Detect l/h transition on CTS */
#define CD1865_MCOR2_CDOD	1<<6	/* Detect l/h transition on CD */
#define CD1865_MCOR2_DSROD	1<<7	/* Detect l/h transition on DSR */

#define CD1865_SRER_NNDT	1<<0	/* Now New Data Timeout */
#define CD1865_SRER_TXMPTY	1<<1	/* Transmitter Empty */
#define CD1865_SRER_TXRDY	1<<2	/* Transmitter Ready Service Request */
#define CD1865_SRER_RXSC	1<<3	/* Receive Special Character Service Request */
#define CD1865_SRER_RDSR	1<<4	/* Receive Data Service Request */
#define CD1865_SRER_CTS		1<<5	/* Clear To Send Service Request */
#define CD1865_SRER_CD		1<<6	/* Carrier Detect (CD) Service Request */
#define CD1865_SRER_DSR		1<<7	/* Data Set Ready Service Request */

#define CD1865_SRCR_PKGTYP	1<<7
#define	CD1865_SRCR_REGACKEN	1<<6	/* Register based Service Request Acknowledge */
#define CD1865_SRCR_DAISYEN	1<<5	/* Daisy Chain RegAckEn */
#define CD1865_SRCR_GLOBPRI	1<<4	/* Global Pri (AuroPri over Multiple CD1865) */
#define CD1865_SRCR_UNFAIR	1<<3	/* Fairness Override Bit */
#define CD1865_SRCR_AUTOPRI	1<<1	/* Auto priritize the service requests */
#define CD1865_SRCR_PRISEL	1<<0	/* Direction of AutoPri */

#define CD1865_RCSR_TIMEOUT	1<<7	/* Receive Timeout */
#define CD1865_RCSR_BREAK	1<<3	/* Break Received */
#define CD1865_RCSR_PARITY	1<<2	/* Parity Error */
#define CD1865_RCSR_FRAME	1<<1	/* Frame Error ? */
#define CD1865_RCSR_OVERRUN	1<<0	/* Receive Fifo Overrun */

#define CD1865_SRSR_RREQ	1<<4
#define CD1865_SRSR_TREQ	1<<2
#define CD1865_SRSR_MREQ	1<<1

#define CD1865_FIFO_DEPTH	8

#define CD1865_C_ESC		0x0
#define CD1865_C_SBRK		0x81
#define CD1865_C_DELAY		0x82
#define CD1865_C_EBRK		0x83

#define AA_MAX_BOARD		4
#define	AA_MAX_CD1865		2	
#define AA_CD1865_PORT		8
#define AA_CCR_TIMEOUT		10000
#define AA_HWRESET_TIMEOUT	10000
#define AA_PORT_BUFFER_SIZE	PAGE_SIZE 

#define AA_MAX_PORT	AA_MAX_BOARD*AA_MAX_CD1865*AA_CD1865_PORT

#define AA_CD1865_FAILED	1<<0

struct aa_board_struct {
	long	plx_base;
	long	port_base;
	int	irq;
	struct  cd1865_struct {
		struct aa_board_struct	*board;
		long			cd1865_base;
		long			state;
		unsigned long		page;			/* A page per card */
		struct semaphore	page_sem;		/* Protect the bounce page */
		spinlock_t		access_lock;		/* Spinlock all instances accessing the channel specific registers */
		struct	port_struct {
			int			port;	/* Port No */

			struct tty_struct	*tty;
			struct cd1865_struct	*cd1865;
			struct wait_queue	*close_wait;
			struct tq_struct        task;
			struct termios		termios;

			int			flags;
			int			count;

			long			pgrp;
			long			session;

			int			break_length;
			
			int			closing_wait;
			int			event;
			/* Chip/Port status register */
			unsigned char		cor1,	/* Channel Option Register 1 (No Bits, Stop Bits, Parity) */
						cor2,	/* Channel Option Register 2 (Flow Control (Hard & Soft) */
						cor3,	/* Channel Option Register 3 (Flow Control & Receive Threshold */
						mcor1,	/* Modem Change Option Register 1 DSR/CD/CTS transition & DTRthreshold */
						mcor2,	/* Modem Change Option Register 2 DSR/CD/CTS transition */
						srer,	/* Service Request Enable Register */
						msvr;	/* Modem Signal Value Register */
			int			bpr;	/* Bit Rate Period Register */
			/* Buffer handling */
			int			xmit_cnt,
						xmit_head,
						xmit_tail,
						wakeup_chars;
			spinlock_t		buffer_lock;	/* Spinlock all instances accessing the channel specific registers */
			unsigned char		*buffer;
		} port[AA_CD1865_PORT];
	} cd1865[AA_MAX_CD1865];
};

